package {
	import com.bit101.components.PushButton;
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.display.BlendMode;
	import flash.display.Shader;
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.KeyboardEvent;
	import flash.events.MouseEvent;
	import flash.filters.ShaderFilter;
	import flash.geom.Matrix3D;
	import flash.geom.Vector3D;
	import flash.ui.Keyboard;
	import net.hires.debug.Stats;
	
	/**
	 * ...
	 * @author Volgogradetzzz
	 */
	public class Main extends Sprite {
		
		private const _CAMERA_ANG:Number = 30 * Math.PI / 180;
		private const _BOUNDS_MIN_X:Number = 0;
		private const _BOUNDS_MAX_X:Number = 500;
		private const _BOUNDS_MIN_Y:Number = 0;
		private const _BOUNDS_MAX_Y:Number = 1000;
		private const _BOUNDS_MIN_Z:Number = 0;
		private const _BOUNDS_MAX_Z:Number = 500;
		
		[Embed(source = "../lib/TrueDepthSort.pbj", mimeType = "application/octet-stream")]
		private const _blendShaderDef:Class;
		private const _blendShader:Shader = new Shader(new _blendShaderDef());
		
		[Embed(source = "../lib/TrueFilter.pbj", mimeType = "application/octet-stream")]
		private var _shaderFilterDef:Class;
		private var _shaderFilter:Shader = new Shader(new _shaderFilterDef());
		private var _filter:ShaderFilter = new ShaderFilter(_shaderFilter);
		
		private var _allObjects:Vector.<IsoObject> = new Vector.<IsoObject>();
		private var _userInput:UserInput = new UserInput();
		private var _selectedObj:IsoObject;
		private var _depthContainer:Sprite = new Sprite();
		private var _depthBG:Bitmap = new Bitmap(new BitmapData(500, 500, true, 0xFF000000));
		private var _depthMap:Bitmap = new Bitmap(new BitmapData(500, 500, true, 0xFF000000));
		private var _shadowsContainer:Sprite = new Sprite();
		private var _objectsContainer:Sprite = new Sprite();
		private var _worldDiagonalProj:Number = 0;
		private var _cameraNormal:Vector3D = new Vector3D(0, 1, 0);
		
		public function Main():void {
			if (stage) init();
			else addEventListener(Event.ADDED_TO_STAGE, init);
		}
		
		private function init(e:Event = null):void {
			removeEventListener(Event.ADDED_TO_STAGE, init);
			
			addChild(new Stats());
			
			addChild(_shadowsContainer);
			addChild(_objectsContainer);
			_depthContainer.addChild(_depthBG);
			
			for (var i:uint = 0; i < 10; i++){
				createRandomObject();
			}
			
			var worldDiagonal:Vector3D = new Vector3D(0, _BOUNDS_MAX_Y, _BOUNDS_MAX_Z);
			var m:Matrix3D = new Matrix3D();
			m.appendRotation(_CAMERA_ANG / Math.PI * 180, Vector3D.X_AXIS);
			_cameraNormal = m.transformVector(_cameraNormal);
			_worldDiagonalProj = worldDiagonal.dotProduct(_cameraNormal);
			
			stage.addEventListener(KeyboardEvent.KEY_DOWN, stageKeyDownHandler);
			stage.addEventListener(KeyboardEvent.KEY_UP, stageKeyUpHandler);
			stage.addEventListener(MouseEvent.CLICK, stageClickHandler);
			
			update(null);
			addEventListener(Event.ENTER_FRAME, update);
			
			var but:PushButton = new PushButton(this, 400, 0, '+1', createRandomObject);
		}
		
		private function createRandomObject(event:MouseEvent = null):void {
			var numObjects:uint = 3;
			var obj:IsoObject;
			
			var rand:Number = Math.random();
			if (rand < 1 * (1 / numObjects)) {
				obj = new Fence1(_CAMERA_ANG, -50, -35, -48);
			}else if (rand < 2 * (1 / numObjects)) {
				obj = new Fence2(_CAMERA_ANG, -50, -35, -48);
			}else if (rand < 3 * (1 / numObjects)) {
				obj = new Cube(_CAMERA_ANG, -50, -15, -50);
			}
			
			obj.x = int(Math.random() * (_BOUNDS_MAX_X - obj.boundingBox.width));
			obj.y = int(Math.random() * (_BOUNDS_MAX_Y - obj.boundingBox.height));
			obj.z = 0;
			
			_objectsContainer.addChild(obj.diffuse);
			_depthContainer.addChild(obj.depth);
			_shadowsContainer.addChild(obj.shadow);
			
			_allObjects.push(obj);
		}
		
		private function update(event:Event):void {
			if (_selectedObj != null) {
				if (_userInput.left) {
					_selectedObj.x -= 3;
				}
				
				if (_userInput.right) {
					_selectedObj.x += 3;
				}
				
				if (_userInput.up) {
					if (_userInput.shift) {
						_selectedObj.z += 3;
					}else {
						_selectedObj.y -= 3;
					}
				}
				
				if (_userInput.down) {
					if (_userInput.shift) {
						_selectedObj.z -= 3;
					}else {
						_selectedObj.y += 3;
					}
				}
				
				checkBounds(_selectedObj);
			}
			
			if (_userInput.leftMouse) {
				if (_selectedObj != null) {
					_selectedObj = null;
				}
				
				var obj:IsoObject = getObjectUnderMouse(_userInput.mouseX, _userInput.mouseY);
				if (obj != null) {
					_selectedObj = obj;
				}
				
				_userInput.leftMouse = false;
			}
			
			applyShaders();
		}
		
		private function applyShaders():void {
			var pos:Vector3D = new Vector3D();
			var posProj:Number = 0;
			
			var objDiagonal:Vector3D = new Vector3D();
			var objDiagonalProj:Number = 0;
			
			for each(var obj:IsoObject in _allObjects) {
				pos.x = obj.x;
				pos.y = obj.y;
				pos.z = obj.z;
				
				posProj = pos.dotProduct(_cameraNormal);
				
				objDiagonal.x = 0;
				objDiagonal.y = obj.boundingBox.depth;
				objDiagonal.z = obj.boundingBox.height;
				
				objDiagonalProj = objDiagonal.dotProduct(_cameraNormal);
				
				_blendShader.data.relZ.value = [posProj / _worldDiagonalProj];
				_blendShader.data.relHeight.value = [objDiagonalProj / _worldDiagonalProj];
				
				obj.depth.blendMode = BlendMode.NORMAL;
				obj.depth.blendShader = _blendShader;
			}
			
			_depthMap.bitmapData.draw(_depthContainer);
			
			for each(obj in _allObjects) {
				pos.x = obj.x;
				pos.y = obj.y;
				pos.z = obj.z;
				
				posProj = pos.dotProduct(_cameraNormal);
				
				objDiagonal.x = 0;
				objDiagonal.y = obj.boundingBox.depth;
				objDiagonal.z = obj.boundingBox.height;
				
				objDiagonalProj = objDiagonal.dotProduct(_cameraNormal);
				
				_filter.shader.data.pos.value[0] = obj.diffuse.x
				_filter.shader.data.pos.value[1] = obj.diffuse.y;
				_filter.shader.data.relZ.value = [posProj / _worldDiagonalProj];
				_filter.shader.data.relHeight.value = [objDiagonalProj / _worldDiagonalProj];
				_filter.shader.data.personalHeightMap.input = obj.depth.bitmapData;
				_filter.shader.data.bufferHeightMap.input =	_depthMap.bitmapData;
				obj.diffuse.filters = [_filter];
			}
		}
		
		private function checkBounds(building:IsoObject):void {
			if (building.x < _BOUNDS_MIN_X) {
				building.x = _BOUNDS_MIN_X;
			}else if (building.x > _BOUNDS_MAX_X - building.boundingBox.width) {
				building.x = _BOUNDS_MAX_X - building.boundingBox.width;
			}
			
			if (building.y < _BOUNDS_MIN_Y) {
				building.y = _BOUNDS_MIN_Y;
			}else if (building.y > _BOUNDS_MAX_Y - building.boundingBox.depth) {
				building.y = _BOUNDS_MAX_Y - building.boundingBox.depth;
			}
			
			if (building.z < _BOUNDS_MIN_Z) {
				building.z = _BOUNDS_MIN_Z;
			}else if (building.z > _BOUNDS_MAX_Z - building.boundingBox.height) {
				building.z = _BOUNDS_MAX_Z - building.boundingBox.height;
			}
		}
		
		private function getObjectUnderMouse(mx:Number, my:Number):IsoObject {
			for each(var obj:IsoObject in _allObjects) {
				if (obj.boundingRect.contains(mx, my)) {
					return obj;
				}
			}
			
			return null;
		}
		
		private function stageKeyDownHandler(event:KeyboardEvent):void {
			if (event.keyCode == Keyboard.A || event.keyCode == Keyboard.LEFT) {
				_userInput.left = true;
			}else if (event.keyCode == Keyboard.D || event.keyCode == Keyboard.RIGHT) {
				_userInput.right = true;
			}else if (event.keyCode == Keyboard.W || event.keyCode == Keyboard.UP) {
				_userInput.up = true;
			}else if (event.keyCode == Keyboard.S || event.keyCode == Keyboard.DOWN) {
				_userInput.down = true;
			}else if (event.keyCode == Keyboard.SHIFT) {
				_userInput.shift = true;
			}
		}
		
		private function stageKeyUpHandler(event:KeyboardEvent):void {
			if (event.keyCode == Keyboard.A || event.keyCode == Keyboard.LEFT) {
				_userInput.left = false;
			}else if (event.keyCode == Keyboard.D || event.keyCode == Keyboard.RIGHT) {
				_userInput.right = false;
			}else if (event.keyCode == Keyboard.W || event.keyCode == Keyboard.UP) {
				_userInput.up = false;
			}else if (event.keyCode == Keyboard.S || event.keyCode == Keyboard.DOWN) {
				_userInput.down = false;
			}else if (event.keyCode == Keyboard.SHIFT) {
				_userInput.shift = false;
			}
		}
		
		private function stageClickHandler(event:MouseEvent):void {
			_userInput.leftMouse = true;
			_userInput.mouseX = stage.mouseX;
			_userInput.mouseY = stage.mouseY;
		}
	}
}